"""Base node module."""

from __future__ import annotations

import dataclasses
import inspect
import re
from typing import Any, Callable

from rdflib import term
from rdflib.namespace import SDO

from mlcroissant._src.core import constants
from mlcroissant._src.core import dataclasses as mlc_dataclasses
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import CroissantVersion
from mlcroissant._src.core.data_types import check_expected_type
from mlcroissant._src.core.issues import Issues
from mlcroissant._src.core.issues import WarningException
from mlcroissant._src.core.json_ld import box_singleton_list
from mlcroissant._src.core.json_ld import remove_empty_values
from mlcroissant._src.core.json_ld import sort_dict
from mlcroissant._src.core.json_ld import unbox_singleton_list
from mlcroissant._src.core.types import Json
from mlcroissant._src.core.uuid import generate_uuid
from mlcroissant._src.core.uuid import uuid_from_jsonld

NAME_REGEX = "[a-zA-Z0-9\\-_\\.]+"
_MAX_NAME_LENGTH = 255
# This could also be an attribute of JsonldField:
_LIST_FIELDS = {"distribution", "fields", "record_sets"}
_MISSING_JSONLD_TYPE = "__MISSING_JSONLD_TYPE__"
MATCHING_TYPES = {
    SDO.Boolean: bool,
    SDO.Date: str,
    SDO.DateTime: str,
    SDO.Integer: int,
    SDO.Number: float,
    SDO.Text: str,
    SDO.Time: str,
    SDO.URL: str,
}


@dataclasses.dataclass(eq=False, repr=False)
class Node:
    """Structure node in Croissant.

    This generic class will be inherited by the actual Croissant nodes:
    - Field
    - FileObject
    - FileSet
    - Metadata
    - RecordSet

    When building the node, `self.context.issues` are populated when issues are
    encountered.

    Args:
        ctx: the context containing the shared state between all nodes.
        name: The name of the node.
        parents: The parent nodes in the Croissant JSON-LD as a tuple.
    """

    ctx: Context = dataclasses.field(default_factory=Context)
    id: str = dataclasses.field(default_factory=generate_uuid)
    name: str | None = None
    parents: list[Node] = dataclasses.field(default_factory=list)
    jsonld: Any = None

    def __post_init__(self):
        """Checks exclusive properties."""
        self._cast_fields()
        self._check_exclusive_properties()

    def _cast_fields(self):
        """Applies all `field.cast_fn`."""
        for field in mlc_dataclasses.jsonld_fields(self):
            if field.cast_fn:
                try:
                    value = getattr(self, field.name)
                    new_value = field.cast_fn(value)
                    setattr(self, field.name, new_value)
                except WarningException as e:
                    self.add_warning(repr(e))
                except Exception as e:
                    self.add_error(repr(e))

    def assert_has_mandatory_properties(self, *mandatory_properties: str):
        """Checks a node in the graph for existing properties with constraints.

        Args:
            mandatory_properties: A list of mandatory properties for the current node.
                If the node doesn't have one, it triggers an error.
        """
        for mandatory_property in mandatory_properties:
            if hasattr(self, mandatory_property):
                value = getattr(self, mandatory_property)
                if not value:
                    error = (
                        "Property"
                        f' "{constants.FROM_CROISSANT(self.ctx).get(mandatory_property)}"'
                        " is mandatory, but does not exist."
                    )
                    self.add_error(error)
            else:
                self.add_error(
                    "mlcroissant checks for an inexisting property:"
                    f" {mandatory_property}"
                )

    def assert_has_optional_properties(self, *optional_properties: str):
        """Checks a node in the graph for existing properties with constraints.

        Args:
            optional_properties: A list of optional properties for the current node. If
                the node doesn't have one, it triggers a warning.
        """
        for optional_property in optional_properties:
            if hasattr(self, optional_property):
                value = getattr(self, optional_property)
                if not value:
                    property = constants.FROM_CROISSANT(self.ctx).get(optional_property)
                    if property is None:
                        self.add_error(
                            "mlcroissant does not define property"
                            f' "{optional_property}" in constants.FROM_CROISSANT.'
                        )
                    else:
                        self.add_warning(
                            f'Property "{property}" is recommended, but does not exist.'
                        )
            else:
                self.add_error(
                    "mlcroissant checks for an inexisting property:"
                    f" {optional_property}"
                )

    def assert_has_exclusive_properties(self, *exclusive_properties: list[str]):
        """Checks a node in the graph for existing properties with constraints.

        Args:
            exclusive_properties: A list of list of exclusive properties: the current
                node should have at least one.
        """
        for possible_exclusive_properties in exclusive_properties:
            if not self.there_exists_at_least_one_property(
                possible_exclusive_properties
            ):
                error = (
                    "At least one of these properties should be defined:"
                    f" {possible_exclusive_properties}."
                )
                self.add_error(error)

    def _check_exclusive_properties(self):
        """Checks exclusive properties on the node."""
        fields = {field.name: field for field in mlc_dataclasses.jsonld_fields(self)}
        for field in mlc_dataclasses.jsonld_fields(self):
            exclusive_with = field.exclusive_with
            if field.name in field.exclusive_with:
                raise ValueError(f'"{field.name}" appears in its own in exclusive_with')
            if field.exclusive_with:
                exclusive_fields = [field] + [
                    fields[name] for name in exclusive_with if name in fields
                ]
                properties = [getattr(self, field.name) for field in exclusive_fields]
                has_exactly_one_property = sum(bool(p) for p in properties) == 1
                if not has_exactly_one_property:
                    urls = [str(field.call_url(self.ctx)) for field in exclusive_fields]
                    name = self._jsonld_type(self.ctx) or self.__class__.__name__
                    self.ctx.issues.add_error(
                        f"{name} should have one of the following properties"
                        f" {' or '.join(urls)}."
                    )

    def get_issue_context(self) -> str:
        """Adds context to an issue by printing metadata(...) > ... > field(...)."""
        nodes = self.parents + [self]
        return " > ".join(f"{type(node).__name__}({node.name})" for node in nodes)

    def add_error(self, error: str):
        """Adds a new error."""
        self.ctx.issues.add_error(error, self)

    def add_warning(self, warning: str):
        """Adds a new warning."""
        self.ctx.issues.add_warning(warning, self)

    def __repr__(self) -> str:
        """Prints a simplified string representation of the node: `Node(uuid="xxx")`."""
        return f'{self.__class__.__name__}(uuid="{self.uuid}")'

    @property
    def uuid(self) -> str:
        """Unique identifier for the node.

        For Croissant <=0.8: for fields, the UID is the path from their RecordSet. For other
        nodes, it is their names.
        For Croissant <=1.0: it corresponds to the node's UUID.
        """
        if self.ctx.is_v0():
            if len(self.parents) <= 1:
                return self.name or ""
            return f"{self.parents[-1].uuid}/{self.name}"
        else:
            return self.id

    @property
    def parent(self) -> Node | None:
        """Direct parent of the node or None if no parent."""
        if not self.parents:
            return None
        return self.parents[-1]

    @property
    def predecessors(self) -> set[Node]:
        """Predecessors in the structure graph."""
        try:
            predecessors = self.ctx.graph.predecessors(self)
            return set(predecessors)  # pytype: disable=bad-return-type
        except KeyError as e:
            raise KeyError(
                f"Could not find node '{self.id}' in the graph. Make sure to build a"
                " full mlcroissant metadata object (mlc.Metadata) wrapping all the"
                " FileSets/FileObjects/RecordSets/Fields."
            ) from e

    @property
    def recursive_predecessors(self) -> set[Node]:
        """Predecessors and predecessors of predecessors in the structure graph."""
        predecessors = set()
        for predecessor in self.predecessors:
            predecessors.add(predecessor)
            predecessors.update(predecessor.recursive_predecessors)
        return predecessors

    @property
    def predecessor(self) -> Node | None:
        """First predecessor in the structure graph."""
        if not self.ctx.graph.has_node(self):
            return None
        return next(
            self.ctx.graph.predecessors(self), None
        )  # pytype: disable=bad-return-type

    @property
    def successors(self) -> tuple[Node, ...]:
        """Successors in the structure graph."""
        if self not in self.ctx.graph:
            return ()
        # We use tuples in order to have a hashable data structure to be put in input of
        # operations.
        return tuple(self.ctx.graph.successors(self))  # pytype: disable=bad-return-type

    @property
    def recursive_successors(self) -> set[Node]:
        """Successors and successors of successors in the structure graph."""
        successors = set()
        for successor in self.successors:
            successors.add(successor)
            successors.update(successor.recursive_successors)
        return successors

    @property
    def successor(self) -> Node | None:
        """Direct successor in the structure graph."""
        if not self.ctx.graph.has_node(self):
            return None
        return next(
            self.ctx.graph.successors(self), None
        )  # pytype: disable=bad-return-type

    @property
    def issues(self) -> Issues:
        """Shortcut to access issues in node."""
        return self.ctx.issues

    def validate_name(self):
        """Validates the name."""
        name = self.name
        if not isinstance(name, str):
            self.add_error(f"The name should be a string. Got: {type(name)}.")
            return
        if not name:
            # This case is already checked for in every node's __post_init__ as `name`
            # is a mandatory parameter for Croissant 0.8
            return
        # For Croissant >= 1.0 compliant datasets, we don't enforce any more constraints
        # on names.
        if not self.ctx.is_v0():
            return
        if len(name) > _MAX_NAME_LENGTH:
            self.add_error(
                f'The name "{name}" is too long (>{_MAX_NAME_LENGTH} characters).'
            )
        regex = re.compile(rf"^{NAME_REGEX}$")
        if not regex.match(name):
            self.add_error(f'The name "{name}" contains forbidden characters.')

    def there_exists_at_least_one_property(self, possible_properties: list[str]):
        """Checks for the existence of one of `possible_properties` in `keys`."""
        for possible_property in possible_properties:
            if getattr(self, possible_property, None) is not None:
                return True
        return False

    def __hash__(self):
        """Re-uses parent's hash function (i.e., object)."""
        return super().__hash__()

    def __eq__(self, other: Any) -> bool:
        """Compares two Nodes given their arguments."""
        if isinstance(other, Node):
            return self.uuid == other.uuid
        return False

    def __deepcopy__(self, memo):
        """Overwrites [`copy.deepcopy`](https://docs.python.org/3/library/copy.html).

        This is required to use `copy.deepcopy(metadata)` as needed by e.g. PyTorch
        DataPipes. It is unsure what the full consequences are. So we'll keep
        investigating on a better approach in
        https://github.com/mlcommons/croissant/issues/531.
        """
        kwargs = {}
        for field in dataclasses.fields(self.__class__):
            if field.name in self.__dict__:
                kwargs[field.name] = self.__dict__[field.name]
        # We properly selected the correct kwargs:
        copy = self.__class__(**kwargs)  # pytype: disable=not-instantiable
        memo[id(self)] = copy
        return copy

    def to_json(self) -> Json:
        """Converts the Python class to JSON."""
        cls = self.__class__
        jsonld_type = cls._jsonld_type(self.ctx)
        # IDs that are generated by RDFLib are not kept
        id_by_rdflib = self.id and self.id.startswith("_:")
        jsonld = {
            "@type": self.ctx.rdf.shorten_value(jsonld_type) if jsonld_type else None,
            "@id": None if self.ctx.is_v0() or id_by_rdflib else self.id,
        }
        for field in mlc_dataclasses.jsonld_fields(self):
            url = field.call_url(self.ctx)
            key = self.ctx.rdf.shorten_key(url)
            value = getattr(self, field.name)
            if field.to_jsonld:
                # We explicitly set `to_jsonld`, so use it:
                value = field.call_to_jsonld(self.ctx, value)
            else:
                if isinstance(value, list):
                    value = [_value_to_jsonld(v) for v in value]
                else:
                    value = _value_to_jsonld(value)
            # fields in _LIST_FIELDS are always lists, so we don't unbox them.
            if field.cardinality == "MANY" and field.name not in _LIST_FIELDS:
                value = unbox_singleton_list(value)
            jsonld[key] = value
        jsonld = remove_empty_values(jsonld)
        return sort_dict(jsonld)

    @classmethod
    def from_jsonld(cls, ctx: Context, jsonld: Json):
        """Creates a Python class from JSON-LD."""
        if not isinstance(jsonld, list) and not isinstance(jsonld, dict):
            name = cls._jsonld_type(ctx) or cls.__name__
            ctx.issues.add_error(f'{name} should be a dict with keys. Got "{jsonld}"')
            return None
        if cls._jsonld_type(ctx) == constants.SCHEMA_ORG_DATASET:
            # For `Metadata` node, insert conforms_to/is_live_dataset in the context:
            ctx.conforms_to = CroissantVersion.from_jsonld(
                ctx, jsonld.get(constants.DCTERMS_CONFORMS_TO)
            )
            ctx.is_live_dataset = jsonld.get(constants.ML_COMMONS_IS_LIVE_DATASET(ctx))
        if isinstance(jsonld, list):
            return [cls.from_jsonld(ctx, el) for el in jsonld]
        check_expected_type(ctx.issues, jsonld, cls._jsonld_type(ctx))
        kwargs = {}
        for field in mlc_dataclasses.jsonld_fields(cls):
            url = field.call_url(ctx)
            value = jsonld.get(url)
            if field.from_jsonld:
                # We explicitly set `from_jsonld`, so use it:
                value = field.call_from_jsonld(ctx, value)
            else:
                # We can infer `from_jsonld` from the input_types:
                if isinstance(value, list):
                    value = [_value_from_input_types(ctx, v, field) for v in value]
                    value = list(filter(lambda v: v != dataclasses.MISSING, value))
                else:
                    value = _value_from_input_types(ctx, value, field)
            if field.cardinality == "MANY" and value:
                value = box_singleton_list(value) if value else []
            elif field.cardinality == "ONE" and isinstance(value, list):
                value = value[0] if value else None
                warning = f"`{field.name}` has cardinality `ONE`, but got a list"
                ctx.issues.add_warning(warning)
            if value is not None:
                kwargs[field.name] = value
        return cls(
            ctx=ctx,
            id=uuid_from_jsonld(jsonld),
            jsonld=jsonld,
            **kwargs,
        )

    JSONLD_TYPE: Callable[[Context], term.URIRef] | term.URIRef | str | None = (
        _MISSING_JSONLD_TYPE
    )

    @classmethod
    def _jsonld_type(cls, ctx: Context):
        """Get the actual JSON-LD type according the the ctx."""
        if cls.JSONLD_TYPE == _MISSING_JSONLD_TYPE:
            raise NotImplementedError("Output the right JSON-LD type.")
        elif callable(cls.JSONLD_TYPE):
            return cls.JSONLD_TYPE(ctx)
        else:
            return cls.JSONLD_TYPE


def _value_to_jsonld(value: Any) -> Any:
    """Applies `to_json` to Nodes."""
    if isinstance(value, Node):
        return value.to_json()
    else:
        return value


def _value_from_input_types(
    ctx: Context, value: Any, field: mlc_dataclasses.JsonldField
) -> Any:
    """Retrieves the value based on the JsonldField."""
    if value is None:
        return None
    input_types = field.input_types
    if not input_types:
        # This is a problem in mlcroissant, so we raise an error:
        raise ValueError(f"`{field.name}` doesn't declare any `input_types`.")
    for input_type in input_types:
        # Either the input_type is a Node...
        if isinstance(value, dict) and _is_a_node(input_type):
            jsonld_type = input_type._jsonld_type(ctx)
            actual_jsonld_type = value.get("@type")
            if actual_jsonld_type == jsonld_type:
                return input_type.from_jsonld(ctx, value)
        # ...or it's a basic int/str/bool/etc type
        else:
            matching_type = MATCHING_TYPES.get(input_type)
            if type(value) is matching_type:
                return value
    types = [
        input_type._jsonld_type(ctx) if _is_a_node(input_type) else str(input_type)
        for input_type in input_types
    ]
    if isinstance(value, dict):
        actual_jsonld_type = value.get("@type")
        posible_attributes = " or ".join([f'"@type": "{type}"' for type in types])
        if ctx.is_v0():
            uuid = value.get(constants.SCHEMA_ORG_NAME)
        else:
            uuid = uuid_from_jsonld(value)
        ctx.issues.add_error(
            f'"{uuid}" should have an attribute {posible_attributes}. Got'
            f" {actual_jsonld_type} instead."
        )
    else:
        actual_type = type(value).__name__
        ctx.issues.add_error(
            f"`{field.name}` should have type {' or '.join(types)}, but got"
            f" {actual_type}"
        )
    if field.default != dataclasses.MISSING:
        # If the validation failed, we return the default value:
        return field.default
    elif field.default_factory != dataclasses.MISSING:
        # For lists, we don't return field.default_factory() to not produce [[]]:
        return dataclasses.MISSING
    raise ValueError(f"{field.name} specifies neither default, nor default_factory")


def _is_a_node(input_type: Any) -> bool:
    return inspect.isclass(input_type) and issubclass(input_type, Node)


def node_by_uuid(ctx: Context, uuid: str | None) -> Node | None:
    """Retrieves a node in the graph by its UID."""
    for node in ctx.graph.nodes():
        if node.uuid == uuid:  # pytype: disable=attribute-error
            return node  # pytype: disable=bad-return-type
    return None



"""constants_test module."""

from mlcroissant._src.core.constants import TO_CROISSANT
from mlcroissant._src.core.context import Context
from mlcroissant._src.tests.versions import parametrize_conforms_to


@parametrize_conforms_to()
def test_to_croissant_values_are_unique(conforms_to):
    ctx = Context(conforms_to=conforms_to)
    deja_vu = {}
    for key, value in TO_CROISSANT(ctx).items():
        if value in deja_vu:
            raise ValueError(
                f"Keys {key} and {deja_vu[value]} define the same Croissant value:"
                f" {value}."
            )
        deja_vu[value] = key



"""download_test module."""

import hashlib
import os
import tempfile

from etils import epath
import pytest

from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import CroissantVersion
from mlcroissant._src.operation_graph.operations.download import _get_hash_algorithm
from mlcroissant._src.operation_graph.operations.download import Download
from mlcroissant._src.operation_graph.operations.download import extract_git_info
from mlcroissant._src.operation_graph.operations.download import get_download_filepath
from mlcroissant._src.operation_graph.operations.download import insert_credentials
from mlcroissant._src.structure_graph.nodes.file_object import FileObject
from mlcroissant._src.structure_graph.nodes.metadata import Metadata
from mlcroissant._src.tests.nodes import create_test_file_object
from mlcroissant._src.tests.nodes import empty_file_object
from mlcroissant._src.tests.operations import operations


def test_str_representation():
    operation = Download(operations=operations(), node=empty_file_object)
    assert str(operation) == "Download(file_object_name)"


def test_extract_git_info():
    with pytest.raises(ValueError, match="unknown git host"):
        extract_git_info("https://thisisnotagithost.com")

    assert extract_git_info("https://github.com/mlcommons/croissant") == (
        "https://github.com/mlcommons/croissant",
        None,
    )
    assert extract_git_info("https://gitlab.com/mlcommons/croissant") == (
        "https://gitlab.com/mlcommons/croissant",
        None,
    )
    assert extract_git_info("https://huggingface.co/datasets/mnist") == (
        "https://huggingface.co/datasets/mnist",
        None,
    )
    assert extract_git_info("https://huggingface.co/datasets/mlcommons/mnist") == (
        "https://huggingface.co/datasets/mlcommons/mnist",
        None,
    )
    assert extract_git_info(
        "https://huggingface.co/datasets/mnist/tree/refs%2Fconvert%2Fparquet"
    ) == ("https://huggingface.co/datasets/mnist", "refs/convert/parquet")
    assert extract_git_info(
        "https://huggingface.co/datasets/mlcommons/mnist/tree/refs%2Fconvert%2Fparquet"
    ) == ("https://huggingface.co/datasets/mlcommons/mnist", "refs/convert/parquet")


def test_insert_credentials():
    assert (
        insert_credentials(
            "https://github.com/mlcommons/croissant", username=None, password=None
        )
        == "https://github.com/mlcommons/croissant"
    )
    assert (
        insert_credentials(
            "https://github.com/mlcommons/croissant",
            username="username@mlcommons.com",
            password="my/password",
        )
        == "https://username%40mlcommons.com:my/password@github.com/mlcommons/croissant"
    )
    with pytest.raises(ValueError, match="provide both"):
        insert_credentials(
            "https://github.com/mlcommons/croissant",
            username="username@mlcommons.com",
            password=None,
        )
    with pytest.raises(ValueError, match="provide both"):
        insert_credentials(
            "https://github.com/mlcommons/croissant",
            username=None,
            password="my/password",
        )


def test_get_hash_obj_md5():
    node = FileObject(
        id="file-object",
        md5="12345",
    )
    hash_algorithm = _get_hash_algorithm(node)

    assert isinstance(hash_algorithm, type(hashlib.md5()))


def test_get_hash_obj_sha256():
    node = FileObject(
        id="file-object",
        sha256="12345",
    )
    hash_algorithm = _get_hash_algorithm(node)

    assert isinstance(hash_algorithm, type(hashlib.sha256()))


def test_get_download_filepath():
    ctx = Context()
    # With mapping
    ctx.mapping = {"file-object": epath.Path("/bar/foo")}
    node = FileObject(
        ctx=ctx, name="foo", id="file-object", content_url="http://foo", sha256="12345"
    )
    assert get_download_filepath(node) == epath.Path("/bar/foo")

    # Without mapping
    ctx.mapping = {}
    node = FileObject(
        ctx=ctx, name="foo", id="file-object", content_url="http://foo", sha256="12345"
    )
    assert os.fspath(get_download_filepath(node)).endswith(
        "download/croissant-0343a8f6b328d44bfe5b69437797bebc36c59c67ac6527fe1f14684142074fff"
    )


def test_hashes_do_not_match():
    with tempfile.NamedTemporaryFile(delete=False) as f:
        filepath = f.name
        ctx = Context(conforms_to=CroissantVersion.V_1_0, folder=epath.Path())
        metadata = Metadata(ctx=ctx, name="bar")
        file_object = create_test_file_object(
            ctx=ctx,
            name="foo",
            id="file-object",
            content_url=os.fspath(filepath),
            # Hash won't match!
            sha256="12345",
        )
        file_object.parents = [metadata]
        download = Download(operations=operations(), node=file_object)
        with pytest.raises(ValueError, match="is not identical with the reference"):
            download()


@pytest.mark.parametrize("conforms_to", CroissantVersion)
# Test the hex and base64 hash values
@pytest.mark.parametrize(
    "hash_value",
    [
        "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
        "47DEQpj8HBSa+/TImW+5JCeuQeRkm5NMpJWZG3hSuFU=",
    ],
)
def test_sha256_hashes_do_match(conforms_to, hash_value):
    with tempfile.NamedTemporaryFile(delete=False) as f:
        filepath = f.name
        ctx = Context(conforms_to=conforms_to, folder=epath.Path())
        metadata = Metadata(ctx=ctx, name="bar")
        file_object = create_test_file_object(
            name="foo",
            id="file-object",
            content_url=os.fspath(filepath),
            # Hash won't match!
            sha256=hash_value,
        )
        file_object.parents = [metadata]
        download = Download(operations=operations(), node=file_object)
        download()


@pytest.fixture()
def dummy_ctx():
    return Context(
        conforms_to=CroissantVersion.V_1_0,
        folder=epath.Path(),
        is_live_dataset=True,
    )


def test_hashes_are_not_checked_for_live_datasets(dummy_ctx):
    with tempfile.NamedTemporaryFile(delete=False) as f:
        filepath = f.name
        metadata = Metadata(ctx=dummy_ctx, name="bar")
        file_object = create_test_file_object(
            ctx=dummy_ctx,
            name="foo",
            id="file-object",
            content_url=os.fspath(filepath),
            # No hash given, no error raised.
            sha256=None,
        )
        file_object.parents = [metadata]
        download = Download(operations=operations(), node=file_object)
        # No error is raised.
        download()


@pytest.mark.parametrize(
    "sha256,md5",
    [("12345", None), (None, "12345")],
)
def test_hashes_are_checked_for_live_datasets(dummy_ctx, sha256, md5):
    with tempfile.NamedTemporaryFile(delete=False) as f:
        filepath = f.name
        metadata = Metadata(ctx=dummy_ctx, name="bar")
        file_object = create_test_file_object(
            ctx=dummy_ctx,
            name="foo",
            content_url=os.fspath(filepath),
            # Hash won't match.
            sha256=sha256,
            md5=md5,
        )
        file_object.parents = [metadata]
        download = Download(operations=operations(), node=file_object)
        with pytest.raises(ValueError, match="is not identical with the reference"):
            download()


@pytest.mark.parametrize("conforms_to", CroissantVersion)
# Test the hex and base64 hash values
@pytest.mark.parametrize(
    "hash_value", ["d41d8cd98f00b204e9800998ecf8427e", "1B2M2Y8AsgTpgAmY7PhCfg=="]
)
def test_md5_hashes_do_match(conforms_to, hash_value):
    with tempfile.NamedTemporaryFile(delete=False) as f:
        filepath = f.name
        ctx = Context(conforms_to=conforms_to, folder=epath.Path())
        metadata = Metadata(ctx=ctx, name="bar")
        file_object = create_test_file_object(
            name="foo",
            id="file-object",
            content_url=os.fspath(filepath),
            # Hash will match!
            md5=hash_value,
        )
        file_object.parents = [metadata]
        download = Download(operations=operations(), node=file_object)
        download()



"""Migrate to use @id for referencing.

For context: https://github.com/mlcommons/croissant/discussions/506.
"""

from mlcroissant._src.core.rdf import make_context


def migrate_distribution(distribution):
    """Add ids to distributions."""
    for d in distribution:
        d["@id"] = d["name"]
        if "containedIn" in d:
            d_id = d["containedIn"]
            if isinstance(d_id, list):
                output_list = [{"@id": d} for d in d_id]
                d["containedIn"] = output_list
            else:
                d["containedIn"] = {"@id": d_id}


def get_distr_uuid(element, distribution):
    """Get the id of a distribution."""
    for d in distribution:
        if element == d["name"]:
            return d["@id"]
    raise KeyError(f"No {element} in current distribution: {distribution}")


def _migrate_field(json_ld, record_set, distribution):
    """Get the id to the different elements of the jsonld."""
    # Add @id to field.
    field_uuid = record_set["@id"] + "/" + json_ld["name"]
    json_ld["@id"] = field_uuid

    if "source" in json_ld:
        source = json_ld["source"]
        if "fileObject" in source:
            d_uuid = get_distr_uuid(source["fileObject"], distribution)
            json_ld["source"]["fileObject"] = {"@id": d_uuid}
        elif "fileSet" in source:
            d_uuid = get_distr_uuid(source["fileSet"], distribution)
            json_ld["source"]["fileSet"] = {"@id": d_uuid}
        elif "field" in source:
            d_uuid = json_ld["source"]["field"]
            json_ld["source"]["field"] = {"@id": d_uuid}
        else:
            raise ValueError("Source should be a field, file_object or a file_set")

    if "references" in json_ld:
        references = json_ld["references"]
        if "fileObject" in references:
            d_uuid = get_distr_uuid(references["fileObject"], distribution)
            json_ld["references"]["fileObject"] = {"@id": d_uuid}
        elif "fileSet" in references:
            d_uuid = get_distr_uuid(references["fileSet"], distribution)
            json_ld["references"]["fileSet"] = {"@id": d_uuid}
        elif "field" in references:
            d_uuid = json_ld["references"]["field"]
            json_ld["references"]["field"] = {"@id": d_uuid}
        else:
            raise ValueError("Reference should be a field, file_object or a file_set")

    return json_ld


def up(json_ld):
    """Up migration that migrates references to use @id."""
    json_copy = json_ld.copy()
    distribution = json_copy.get("distribution", [])
    json_ld["@context"] = make_context()

    # Add @id to distribution.
    migrate_distribution(distribution)

    for i, record_set in enumerate(json_copy.get("recordSet", [])):
        fields = record_set.get("field", [])
        if isinstance(fields, dict):
            fields = [fields]
        assert isinstance(fields, list), "RecordSet.field is not a list"

    for i, record_set in enumerate(json_copy.get("recordSet", [])):
        # Add @id to recordSet
        record_set["@id"] = record_set["name"]
        fields = record_set.get("field", [])
        if isinstance(fields, dict):
            fields = [fields]
        assert isinstance(fields, list), "RecordSet.field is not a list"
        for j, field in enumerate(fields.copy()):
            field = _migrate_field(field, record_set, distribution)
            sub_fields = field.get("subField", [])
            if isinstance(sub_fields, dict):
                sub_fields = [sub_fields]
            for k, sub_field in enumerate(sub_fields.copy()):
                sub_field = _migrate_field(sub_field, record_set, distribution)
                field["subField"][k] = sub_field
            record_set["field"][j] = field
        json_ld["recordSet"][i] = record_set

    return json_ld



"""read_test module."""

import pathlib
import pickle
import tempfile
from unittest import mock

from etils import epath
import pandas as pd
import pytest

from mlcroissant._src.core.path import Path
from mlcroissant._src.operation_graph.operations.read import _reading_method
from mlcroissant._src.operation_graph.operations.read import Read
from mlcroissant._src.operation_graph.operations.read import ReadingMethod
from mlcroissant._src.structure_graph.nodes.source import Extract
from mlcroissant._src.structure_graph.nodes.source import FileProperty
from mlcroissant._src.structure_graph.nodes.source import Source
from mlcroissant._src.tests.nodes import create_test_field
from mlcroissant._src.tests.nodes import create_test_file_object
from mlcroissant._src.tests.nodes import empty_file_object
from mlcroissant._src.tests.operations import operations


def test_str_representation():
    operation = Read(
        operations=operations(),
        node=empty_file_object,
        folder=epath.Path(),
        fields=(),
    )
    assert str(operation) == "Read(file_object_name)"


def test_explicit_message_when_pyarrow_is_not_installed():
    with mock.patch.object(pd, "read_parquet", side_effect=ImportError):
        with tempfile.TemporaryDirectory() as folder:
            content_url = "file.parquet"
            folder = epath.Path(folder)
            # Create filepath = `folder/file.parquet`.
            filepath = folder / content_url
            file = Path(filepath=filepath, fullpath=pathlib.PurePath())
            filepath.touch()
            read = Read(
                operations=operations(),
                node=create_test_file_object(
                    encoding_format="application/x-parquet", content_url=content_url
                ),
                folder=folder,
                fields=(),
            )
            with pytest.raises(
                ImportError, match=".*pip install mlcroissant\\[parquet\\].*"
            ):
                read.call([file])


def test_reading_method():
    json_field = create_test_field(source=Source(extract=Extract(json_path="path")))
    column_field = create_test_field(source=Source(extract=Extract(column="column")))
    content_field = create_test_field(
        source=Source(extract=Extract(file_property=FileProperty.content))
    )
    lines_field = create_test_field(
        source=Source(extract=Extract(file_property=FileProperty.lines))
    )
    filename = create_test_field(
        source=Source(extract=Extract(file_property=FileProperty.filename))
    )
    assert (
        _reading_method(empty_file_object, (json_field, filename)) == ReadingMethod.JSON
    )
    assert (
        _reading_method(empty_file_object, (column_field, filename))
        == ReadingMethod.CONTENT
    )
    assert (
        _reading_method(empty_file_object, (content_field, filename))
        == ReadingMethod.CONTENT
    )
    assert (
        _reading_method(empty_file_object, (lines_field, filename))
        == ReadingMethod.LINES
    )
    assert _reading_method(empty_file_object, (filename,)) == ReadingMethod.NONE
    with pytest.raises(ValueError):
        _reading_method(empty_file_object, (content_field, lines_field))


def test_pickable():
    operation = Read(
        operations=operations(),
        node=empty_file_object,
        folder=epath.Path("/foo/bar"),
        fields=(),
    )
    operation = pickle.loads(pickle.dumps(operation))
    assert operation.folder == epath.Path("/foo/bar")



"""Migration: https://github.com/mlcommons/croissant/discussions/151."""


def _migrate_field(json_ld):
    for field in ["source", "references"]:
        if field in json_ld:
            source = json_ld[field]
            for old_name, new_name in [
                ("dataExtraction", "extract"),
                ("applyTransform", "transform"),
            ]:
                if old_name in source:
                    source[new_name] = source[old_name]
                    del source[old_name]
            # Sort keys
            source = {k: v for k, v in sorted(source.items())}
            json_ld[field] = source
    return json_ld


def up(json_ld):
    """Up migration that converts `containedIn` to the new format."""
    json_copy = json_ld.copy()
    for i, record_set in enumerate(json_copy.get("recordSet", [])):
        fields = record_set.get("field", [])
        if isinstance(fields, dict):
            fields = [fields]
        for j, field in enumerate(fields):
            field = _migrate_field(field)
            sub_fields = field.get("subField", [])
            if isinstance(sub_fields, dict):
                sub_fields = [sub_fields]
            for k, sub_field in enumerate(sub_fields):
                sub_field = _migrate_field(sub_field)
                field["subField"][k] = sub_field
            record_set["field"][j] = field
        json_ld["recordSet"][i] = record_set
    json_ld["@context"]["extract"] = "ml:extract"
    json_ld["@context"]["transform"] = "ml:transform"
    if "applyTransform" in json_ld["@context"]:
        del json_ld["@context"]["applyTransform"]
    if "dataExtraction" in json_ld["@context"]:
        del json_ld["@context"]["dataExtraction"]
    return json_ld



"""Core utils to manipulate JSON-LD.

The main functions are:
- `expand_json_ld`: human-readable JSON-LD   -> machine-readable JSON-LD.
- `reduce_json_ld`: machine-readable JSON-LD -> human-readable JSON-LD.
"""

import json
import re
from typing import Any

import rdflib
from rdflib import namespace
from rdflib import plugin
from rdflib import term

from mlcroissant._src.core.issues import ValidationError

# This is for compatibility with older versions of rdflib/rdflib-jsonld.
# Indeed, rdflib-jsonld was merged into rdflib from the version 6.0.1.
if rdflib.__version__ < "6.0.1":
    plugin.register(
        "json-ld", plugin.Serializer, "rdflib_jsonld.serializer", "JsonLDSerializer"
    )
    plugin.register("json-ld", plugin.Parser, "rdflib_jsonld.parser", "JsonLDParser")

from mlcroissant._src.core import constants
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import CroissantVersion
from mlcroissant._src.core.rdf import get_context
from mlcroissant._src.core.rdf import make_context
from mlcroissant._src.core.types import Json

_ID_REGEX = re.compile(r".*\s+.*")
_DCTERMS_PREFIX = constants.DCTERMS
_SCHEMA_ORG_PREFIX = constants.SCHEMA_ORG
_WD_PREFIX = constants.WIKIDATA
# Mapping for non-trivial conversion:
_DATA = set()
for conforms_to in CroissantVersion:
    ctx = Context(conforms_to=conforms_to)
    _DATA.add(constants.ML_COMMONS_FIELD_TYPE(ctx))
# Mapping for non-trivial conversion:
_PREFIX_MAP = {}
for conforms_to in CroissantVersion:
    ctx = Context(conforms_to=conforms_to)
    _PREFIX_MAP[constants.ML_COMMONS_FIELD_TYPE(ctx)] = "field"
    _PREFIX_MAP[constants.ML_COMMONS_RECORD_SET_TYPE(ctx)] = "recordSet"
    _PREFIX_MAP[constants.ML_COMMONS_SUB_FIELD_TYPE(ctx)] = "subField"
# List of key/type where `key` always outputs lists when used in nodes of type `type`.
_KEYS_WITH_LIST = set()
for conforms_to in CroissantVersion:
    ctx = Context(conforms_to=conforms_to)
    _KEYS_WITH_LIST.add(
        (constants.ML_COMMONS_FIELD(ctx), constants.ML_COMMONS_RECORD_SET_TYPE(ctx))
    )
    _KEYS_WITH_LIST.add(
        (constants.ML_COMMONS_RECORD_SET(ctx), constants.SCHEMA_ORG_DATASET)
    )
    _KEYS_WITH_LIST.add(
        (constants.ML_COMMONS_SUB_FIELD(ctx), constants.ML_COMMONS_FIELD_TYPE(ctx))
    )
    _KEYS_WITH_LIST.add(
        (constants.SCHEMA_ORG_DISTRIBUTION, constants.SCHEMA_ORG_DATASET)
    )


def _is_dataset_node(node: Json) -> bool:
    """Checks if the type of a node is schema.org/Dataset."""
    node_types = node.get("@type")
    return isinstance(node_types, list) and any(
        node_type == str(constants.SCHEMA_ORG_DATASET) for node_type in node_types
    )


def _sort_items(jsonld: Json) -> list[tuple[str, Any]]:
    """Sorts items from dict.items().

    For human readability, we want "@type"/"name"/"description/conformsTo" to be
    at the beginning of the JSON, while long lists (""distribution"/"recordSet")
    are at the end.
    """
    items = sorted(jsonld.items())
    start_keys = ["@context", "@type", "@id", "name", "description", "conformsTo"]
    end_keys = [
        "distribution",
        "field",
        "data",
        "recordSet",
        "subField",
        "extract",
        "transform",
    ]
    sorted_items = []
    for key in start_keys:
        if key in jsonld:
            sorted_items.append((key, jsonld[key]))
    for item in items:
        if item[0] not in start_keys and item[0] not in end_keys:
            sorted_items.append(item)
    for key in end_keys:
        if key in jsonld:
            sorted_items.append((key, jsonld[key]))
    return sorted_items


def sort_dict(d: Json):
    """Sorts the keys of a nested dict."""
    return {
        k: sort_dict(v) if isinstance(v, dict) and k != "@context" else v
        for k, v in _sort_items(d)
    }


def remove_empty_values(d: Json) -> Json:
    """Removes empty values in a JSON."""
    # Either v exists or v is a bool (to account for False values)
    return {k: v for k, v in d.items() if v or isinstance(v, bool)}


def unbox_singleton_list(d: Any):
    """Returns the first element if one element, else the whole list.

    This is interesting in JSON-LD for properties with a MANY cardinality where we want
    to return one element if there's only one element, else the whole list.
    """
    if isinstance(d, list):
        if len(d) == 1:
            return d[0]
    return d


def box_singleton_list(element: Any) -> list[Any] | None:
    """Returns the element always boxed inside of a list.

    This is interesting for JSON-LD properties with a MANY cardinality (e.g.,
    https://schema.org/containedIn) where we want to return a list even if the input
    element was a singleton.
    """
    if element is None:
        return None
    elif isinstance(element, list):
        return element
    else:
        return [element]


def recursively_populate_jsonld(entry_node: Json, id_to_node: dict[str, Json]) -> Any:
    """Changes in place `entry_node` with its children."""
    if "@value" in entry_node:
        if entry_node.get("@type") == namespace.RDF.JSON:
            # Stringified JSON is loaded as a dict.
            return json.loads(entry_node["@value"])
        else:
            # Other values are loaded as is.
            return entry_node["@value"]
    elif len(entry_node) == 1 and "@id" in entry_node:
        node_id = entry_node["@id"]
        if node_id in id_to_node:
            entry_node = id_to_node[node_id]
            return recursively_populate_jsonld(entry_node, id_to_node)
        else:
            return entry_node
    for key, value in entry_node.copy().items():
        if key == "@type" and isinstance(value, list):
            entry_node[key] = term.URIRef(value[0])
        elif isinstance(value, list):
            del entry_node[key]
            value = [recursively_populate_jsonld(child, id_to_node) for child in value]
            node_type = entry_node.get("@type", "")
            key, node_type = term.URIRef(key), term.URIRef(node_type)
            if (key, node_type) in _KEYS_WITH_LIST:
                entry_node[key] = value
            elif len(value) == 1:
                entry_node[key] = value[0]
            else:
                entry_node[key] = value
    return entry_node


def check_valid_ids(data: Json, ctx: Context) -> None:
    """Checks that the given json contains valid `@id`s."""
    if isinstance(data, dict):
        for k, v in data.items():
            if k == "@id" and re.match(_ID_REGEX, v):
                ctx.issues.add_error(
                    f"The dataset contains a wrong `@id`: '{v}'. Note that currently we"
                    " do not support `@id`s containing whitespaces (not even if"
                    " URL-escaped)."
                )
            if isinstance(v, dict):
                check_valid_ids(v, ctx)
            elif isinstance(v, list):
                for d in v:
                    check_valid_ids(d, ctx)


def expand_jsonld(data: Json, ctx: Context) -> Json:
    """Expands a Croissant JSON to a nested JSON-LD with expanded.

    For this we use RDFLib. RDFLib expands the CURIE of the form "rdf:type" into their
    full expression, but RDFLib also flattens the JSON-LD in a list of nodes. We then
    need to reconstruct the hierarchy.
    """
    check_valid_ids(data=data, ctx=ctx)
    context = get_context(data)
    if "@base" not in context:
        context["@base"] = constants.BASE_IRI
    graph = rdflib.Graph()
    graph.parse(
        data=json.dumps(data),
        format="json-ld",
    )
    # `graph.serialize` outputs a stringified list of JSON-LD nodes.
    nodes = graph.serialize(format="json-ld")
    nodes = json.loads(nodes)
    assert nodes, "Found no node in graph"
    # Find the entry node (schema.org/Dataset). If None found, will raise an error.
    entry_node = next((record for record in nodes if _is_dataset_node(record)), None)
    if entry_node is None:
        ctx.issues.add_error(
            "The current JSON-LD doesn't extend https://schema.org/Dataset."
        )
        raise ValidationError(ctx.issues.report())
    id_to_node: dict[str, Json] = {}
    for node in nodes:
        node_id = node.get("@id")
        id_to_node[node_id] = node
    recursively_populate_jsonld(entry_node, id_to_node)
    entry_node["@context"] = make_context(**context)
    return entry_node


def compact_jsonld(json_: Any) -> Any:
    """Recursively compacts the JSON-LD value to human-readable values.

    For example: "http://schema.org/Dataset" -> "sc:Dataset".
    """
    if isinstance(json_, list):
        return [compact_jsonld(element) for element in json_]
    elif not isinstance(json_, dict):
        if isinstance(json_, str) and _SCHEMA_ORG_PREFIX in json_:
            return json_.replace(_SCHEMA_ORG_PREFIX, "sc:")
        elif isinstance(json_, str) and constants.ML_COMMONS_V_0_8 in json_:
            return json_.replace(constants.ML_COMMONS_V_0_8, "ml:")
        elif isinstance(json_, str) and constants.ML_COMMONS_V_1_0 in json_:
            return json_.replace(constants.ML_COMMONS_V_1_0, "cr:")
        elif isinstance(json_, str) and _DCTERMS_PREFIX in json_:
            return json_.replace(_DCTERMS_PREFIX, "dct:")
        elif isinstance(json_, str) and _WD_PREFIX in json_:
            return json_.replace(_WD_PREFIX, "wd:")
        else:
            return json_
    for key, value in json_.copy().items():
        if key == "@context":
            # `@context` is left untouched.
            continue
        new_value = compact_jsonld(value)
        del json_[key]
        if key == "@id":
            if (
                value.startswith(_SCHEMA_ORG_PREFIX)
                or value.startswith(_DCTERMS_PREFIX)
                or value.startswith(constants.ML_COMMONS_V_0_8)
                or value.startswith(constants.ML_COMMONS_V_1_0)
                or value.startswith(_WD_PREFIX)
            ):
                return new_value
        elif key in _DATA:
            json_["data"] = json.loads(value)
        elif key in _PREFIX_MAP:
            json_[_PREFIX_MAP[key]] = new_value
        elif _SCHEMA_ORG_PREFIX in key:
            new_key = key.replace(_SCHEMA_ORG_PREFIX, "")
            json_[new_key] = new_value
        elif constants.ML_COMMONS_V_0_8 in key:
            new_key = key.replace(constants.ML_COMMONS_V_0_8, "")
            json_[new_key] = new_value
        elif constants.ML_COMMONS_V_1_0 in key:
            new_key = key.replace(constants.ML_COMMONS_V_1_0, "")
            json_[new_key] = new_value
        elif _DCTERMS_PREFIX in key:
            new_key = key.replace(_DCTERMS_PREFIX, "")
            json_[new_key] = value
        elif _WD_PREFIX in key:
            new_key = key.replace(_WD_PREFIX, "")
            json_[new_key] = value
        else:
            json_[key] = new_value
    return sort_dict(json_)



"""Migration: to add the conformsTo field."""

NEW_CONTEXT = {
    "@language": "en",
    "@vocab": "https://schema.org/",
    "column": "ml:column",
    "conformsTo": "dct:conformsTo",
    "data": {"@id": "ml:data", "@type": "@json"},
    "dataBiases": "ml:dataBiases",
    "dataCollection": "ml:dataCollection",
    "dataType": {"@id": "ml:dataType", "@type": "@vocab"},
    "dct": "http://purl.org/dc/terms/",
    "extract": "ml:extract",
    "field": "ml:field",
    "fileProperty": "ml:fileProperty",
    "format": "ml:format",
    "includes": "ml:includes",
    "isEnumeration": "ml:isEnumeration",
    "jsonPath": "ml:jsonPath",
    "ml": "http://mlcommons.org/schema/",
    "parentField": "ml:parentField",
    "path": "ml:path",
    "personalSensitiveInformation": "ml:personalSensitiveInformation",
    "recordSet": "ml:recordSet",
    "references": "ml:references",
    "regex": "ml:regex",
    "repeated": "ml:repeated",
    "replace": "ml:replace",
    "sc": "https://schema.org/",
    "separator": "ml:separator",
    "source": "ml:source",
    "subField": "ml:subField",
    "transform": "ml:transform",
}


def up(json_ld):
    """Up migration to set conformsTo to croissant specs 1.0."""
    json_ld["@context"] = NEW_CONTEXT
    json_ld["conformsTo"] = "http://mlcommons.org/croissant/1.0"
    return json_ld



"""source_test module."""

import pytest
from rdflib.term import URIRef

from mlcroissant._src.core import constants
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import CroissantVersion
from mlcroissant._src.core.issues import Issues
from mlcroissant._src.structure_graph.nodes.source import Extract
from mlcroissant._src.structure_graph.nodes.source import FileProperty
from mlcroissant._src.structure_graph.nodes.source import is_file_property
from mlcroissant._src.structure_graph.nodes.source import Source
from mlcroissant._src.structure_graph.nodes.source import Transform
from mlcroissant._src.tests.nodes import assert_contain_error
from mlcroissant._src.tests.versions import parametrize_conforms_to


def test_source_bool():
    empty_source = Source(transforms=[Transform(replace="\\n/<eos>")])
    assert not empty_source

    whole_source = Source(field="one/two")
    assert whole_source


@pytest.mark.parametrize(
    ["conforms_to", "json_ld", "expected_source"],
    [
        [
            CroissantVersion.V_0_8,
            {
                "@id": "source",
                constants.SCHEMA_ORG_DISTRIBUTION: "my-csv",
                URIRef("http://mlcommons.org/schema/transform"): [
                    {URIRef("http://mlcommons.org/schema/replace"): "\\n/<eos>"},
                    {URIRef("http://mlcommons.org/schema/separator"): " "},
                ],
                URIRef("http://mlcommons.org/schema/extract"): {
                    URIRef("http://mlcommons.org/schema/column"): "my-column"
                },
            },
            Source(
                extract=Extract(column="my-column"),
                id="source",
                transforms=[Transform(replace="\\n/<eos>"), Transform(separator=" ")],
                distribution="my-csv",
            ),
        ],
        [
            CroissantVersion.V_1_0,
            {
                "@id": "source",
                URIRef("http://mlcommons.org/croissant/fileObject"): "my-csv",
                URIRef("http://mlcommons.org/croissant/transform"): [
                    {URIRef("http://mlcommons.org/croissant/replace"): "\\n/<eos>"},
                    {URIRef("http://mlcommons.org/croissant/separator"): " "},
                ],
                URIRef("http://mlcommons.org/croissant/extract"): {
                    URIRef("http://mlcommons.org/croissant/column"): "my-column"
                },
            },
            Source(
                extract=Extract(column="my-column"),
                id="source",
                transforms=[Transform(replace="\\n/<eos>"), Transform(separator=" ")],
                file_object="my-csv",
            ),
        ],
    ],
)
def test_source_parses_list(conforms_to, json_ld, expected_source):
    ctx = Context(conforms_to=conforms_to)
    expected_source.ctx = ctx
    assert Source.from_jsonld(ctx, json_ld) == expected_source
    assert not ctx.issues.errors
    assert not ctx.issues.warnings


@parametrize_conforms_to()
@pytest.mark.parametrize(
    ["jsonld", "expected_error"],
    [
        [
            "this is not a list",
            "Transform should be a dict with keys",
        ],
        [
            [{"not": "the right keys"}],
            "Transform should have one of the following properties",
        ],
    ],
)
def test_transformations_with_errors(conforms_to, jsonld, expected_error):
    ctx = Context(conforms_to=conforms_to)
    Transform.from_jsonld(ctx=ctx, jsonld=jsonld)
    assert len(ctx.issues.errors) == 1
    assert_contain_error(ctx.issues, expected_error)


@pytest.mark.parametrize(
    ["conforms_to", "json_ld"],
    [
        [
            CroissantVersion.V_0_8,
            {
                constants.SCHEMA_ORG_DISTRIBUTION: "my-csv",
                URIRef("http://mlcommons.org/schema/field"): "my-record-set/my-field",
            },
        ],
        [
            CroissantVersion.V_1_0,
            {
                URIRef("http://mlcommons.org/croissant/fileObject"): "my-csv",
                URIRef(
                    "http://mlcommons.org/croissant/field"
                ): "my-record-set/my-field",
            },
        ],
        [
            CroissantVersion.V_1_0,
            {
                URIRef("http://mlcommons.org/croissant/fileSet"): "my-csv",
                URIRef(
                    "http://mlcommons.org/croissant/field"
                ): "my-record-set/my-field",
            },
        ],
        [
            CroissantVersion.V_1_0,
            {
                URIRef("http://mlcommons.org/croissant/fileObject"): "my-csv",
                URIRef("http://mlcommons.org/croissant/fileSet"): "my-csv",
            },
        ],
    ],
)
def test_declaring_multiple_sources_in_one(conforms_to, json_ld):
    ctx = Context(conforms_to=conforms_to)
    Source.from_jsonld(ctx, json_ld)
    assert len(ctx.issues.errors) == 1
    assert_contain_error(ctx.issues, "Source should have one of the following properti")


@parametrize_conforms_to()
def test_declaring_multiple_data_extraction_in_one(conforms_to):
    ctx = Context(conforms_to=conforms_to)
    json_ld = {
        constants.ML_COMMONS_EXTRACT(ctx): {
            "@id": "jsonld-id",
            constants.ML_COMMONS_COLUMN(ctx): "csv_column",
            constants.ML_COMMONS_JSON_PATH(ctx): "json_path",
        },
    }
    assert Source.from_jsonld(ctx, json_ld) == Source(
        extract=Extract(column="csv_column", json_path="json_path"),
    )
    assert len(ctx.issues.errors) == 2
    assert_contain_error(
        ctx.issues, "Source should have one of the following properties"
    )


@parametrize_conforms_to()
def test_declaring_wrong_file_property(conforms_to):
    ctx = Context(conforms_to=conforms_to)
    json_ld = {
        constants.ML_COMMONS_EXTRACT(ctx): {
            constants.ML_COMMONS_FILE_PROPERTY(ctx): "foo",
        },
    }
    Source.from_jsonld(ctx, json_ld)
    assert_contain_error(
        ctx.issues,
        "fileProperty can only have values in"
        " `fullpath`, `filepath` and `content`. Got: foo",
    )


@pytest.mark.parametrize(
    ["source", "expected_column"],
    [
        [
            Source(file_object="my-csv", extract=Extract(column="my-csv-column")),
            "my-csv-column",
        ],
        [
            Source(
                file_object="my-csv",
                extract=Extract(file_property=FileProperty.content),
            ),
            FileProperty.content,
        ],
        [
            Source(
                file_object="my-csv",
                extract=Extract(json_path="/some/json/path"),
            ),
            "/some/json/path",
        ],
        [
            Source(field="record_set/field_name"),
            "record_set/field_name",
        ],
    ],
)
def test_get_field(source: Source, expected_column: str):
    assert source.get_column() == expected_column


def test_is_file_property():
    assert is_file_property("content")
    assert is_file_property("filename")
    assert is_file_property("filepath")
    assert is_file_property("fullpath")
    assert not is_file_property("foo")


def test_check_source_for_valid_json_path():
    issues = Issues()
    Source(id="id", extract=Extract(json_path="*.first.second")).check_source(
        issues.add_error
    )
    assert not issues.errors


def test_check_source_for_invalid_json_path():
    issues = Issues()
    Source(id="id", extract=Extract(json_path="invalid/json/path")).check_source(
        issues.add_error
    )
    errors = list(issues.errors)
    assert len(errors) == 1
    assert "Wrong JSONPath" in errors[0]



"""Tests for optional deps."""

import importlib
from unittest import mock

import pytest

from mlcroissant._src.core.optional import _try_import
from mlcroissant._src.core.optional import deps


@pytest.mark.parametrize(
    "optional_deps",
    [
        "git",
    ],
)
def test_load_optional_deps(optional_deps):
    getattr(deps, optional_deps)


def test_explicit_error_message():
    with mock.patch.object(importlib, "import_module", side_effect=ImportError):
        with pytest.raises(ImportError, match="pip install foo"):
            _try_import("foo")
        with pytest.raises(ImportError, match="pip install bar"):
            _try_import("foo", package_name="bar")



"""PyTorch utilities public API."""

from mlcroissant._src.torch.torch_adapter import LoaderFactory
from mlcroissant._src.torch.torch_adapter import LoaderSpecificationDataType

__all__ = ["LoaderFactory", "LoaderSpecificationDataType"]



"""Generates the dataset and yields the first example."""

import datetime
from typing import TypedDict

from absl import app
from absl import flags
from absl import logging
from etils import epath
from rdflib import term

import mlcroissant as mlc
from mlcroissant._src.core import dataclasses as mlc_dataclasses

flags.DEFINE_string(
    "output",
    None,
    "The output folder.",
    required=True,
)

flags.mark_flag_as_required("output")


FLAGS = flags.FLAGS


class Url(TypedDict):
    """TypedDict with `url` and `name`."""

    url: str
    name: str


def main(argv):
    """Main function launched by the script."""
    del argv
    output = epath.Path(FLAGS.output)
    return documentation(output=output)


def documentation(output: epath.Path):
    """Creates the documentation."""
    if output.exists():
        logging.info(f"Path {output} already exists")
        output = output / datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    output.mkdir(parents=True)
    import jinja2

    loader = jinja2.FileSystemLoader(epath.Path(__file__).parent / "templates")
    env = jinja2.Environment(loader=loader, autoescape=jinja2.select_autoescape())
    ctx = mlc.Context()
    for cls in [mlc.Field, mlc.FileObject, mlc.FileSet, mlc.Metadata, mlc.RecordSet]:
        fields = []
        for field in mlc_dataclasses.jsonld_fields(cls):
            url = field.call_url(ctx)
            url = Url(url=url, name=ctx.rdf.shorten_value(url))
            input_types = []
            for input_type in field.input_types:
                if isinstance(input_type, term.URIRef):
                    input_types.append(
                        Url(url=str(input_type), name=ctx.rdf.shorten_value(input_type))
                    )
                elif issubclass(input_type, mlc.Node):
                    jsonld_type = input_type._jsonld_type(ctx)
                    if isinstance(jsonld_type, term.URIRef):
                        input_types.append(
                            Url(
                                url=str(jsonld_type),
                                name=ctx.rdf.shorten_value(jsonld_type),
                            )
                        )
                    else:
                        input_types.append(Url(url="", name=input_type.__name__))
                else:
                    input_types.append(Url(url=str(input_type), name=""))
            details = {
                "cardinality": field.cardinality,
                "description": field.description,
                "input_types": input_types,
                "url": url,
            }
            fields.append(details)
        data = {
            "title": cls.__name__,
            "documentation": cls.__doc__,
            "fields": fields,
        }
        template = env.get_template("node.html")
        file = output / f"{cls.__name__}.html"
        file.write_text(template.render(data))
    print(f"Wrote documentation to {output}")


if __name__ == "__main__":
    app.run(main)



"""Migration: Remove `distribution` from DataSource, and use `fileObject`/`fileSet`."""


def _migrate_field(json_ld, distribution):
    for field in ["source", "references"]:
        if field in json_ld:
            source = json_ld[field]
            if "distribution" in source:
                resource_name = source["distribution"]
                for resource in distribution:
                    if resource["name"] == resource_name:
                        if resource["@type"] == "sc:FileObject":
                            source["fileObject"] = resource_name
                        elif resource["@type"] == "sc:FileSet":
                            source["fileSet"] = resource_name
                        else:
                            raise ValueError(f"invalid source: {source}")
                del source["distribution"]
                json_ld[field] = source
    return json_ld


def up(json_ld):
    """Up migration that migrates `distribution`->`fileObject`/`fileSet`."""
    json_copy = json_ld.copy()
    distribution = json_copy.get("distribution", [])
    for i, record_set in enumerate(json_copy.get("recordSet", [])):
        fields = record_set.get("field", [])
        if isinstance(fields, dict):
            fields = [fields]
        assert isinstance(fields, list), "RecordSet.field is not a list"
        for j, field in enumerate(fields.copy()):
            field = _migrate_field(field, distribution)
            sub_fields = field.get("subField", [])
            if isinstance(sub_fields, dict):
                sub_fields = [sub_fields]
            for k, sub_field in enumerate(sub_fields.copy()):
                sub_field = _migrate_field(sub_field, distribution)
                field["subField"][k] = sub_field
            record_set["field"][j] = field
        json_ld["recordSet"][i] = record_set
    json_ld["@context"]["fileObject"] = "ml:fileObject"
    json_ld["@context"]["fileSet"] = "ml:fileSet"
    return json_ld



"""Module to manipulate URLs."""

from typing import Any


def is_url(url: Any) -> bool:
    """Tests whether a URL is valid.

    The current version is very loose and only supports the HTTP protocol.
    """
    if isinstance(url, str):
        return url.startswith("http://") or url.startswith("https://")
    return False



"""Tests for FileSets."""

from unittest import mock

import pytest

from mlcroissant._src.core import constants
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import CroissantVersion
from mlcroissant._src.core.uuid import formatted_uuid_to_json
from mlcroissant._src.structure_graph.base_node import Node
from mlcroissant._src.structure_graph.nodes.file_set import FileSet
from mlcroissant._src.tests.nodes import create_test_node
from mlcroissant._src.tests.versions import parametrize_conforms_to


@pytest.mark.parametrize(
    ["conforms_to", "field_uuid"],
    [[CroissantVersion.V_0_8, "name"], [CroissantVersion.V_1_0, "id"]],
)
def test_checks_are_performed(conforms_to, field_uuid):
    with mock.patch.object(
        Node, "assert_has_mandatory_properties"
    ) as mandatory_mock, mock.patch.object(
        Node, "assert_has_optional_properties"
    ) as optional_mock, mock.patch.object(
        Node, "validate_name"
    ) as validate_name_mock:
        ctx = Context(conforms_to=conforms_to)
        create_test_node(FileSet, ctx=ctx)
        mandatory_mock.assert_called_once_with(
            "includes", "encoding_format", field_uuid
        )
        optional_mock.assert_not_called()
        validate_name_mock.assert_called_once()


@parametrize_conforms_to()
def test_from_jsonld(conforms_to):
    ctx = Context(conforms_to=conforms_to)
    contained_in = formatted_uuid_to_json(ctx=ctx, uuid="some.zip")
    jsonld = {
        "@type": constants.SCHEMA_ORG_FILE_SET(ctx),
        "@id": "foo_id",
        constants.SCHEMA_ORG_NAME: "foo",
        constants.SCHEMA_ORG_DESCRIPTION: "bar",
        constants.SCHEMA_ORG_CONTAINED_IN: contained_in,
        constants.SCHEMA_ORG_ENCODING_FORMAT: "application/json",
        constants.ML_COMMONS_EXCLUDES(ctx): "*.csv",
        constants.ML_COMMONS_INCLUDES(ctx): "*.json",
    }
    file_set = FileSet.from_jsonld(ctx, jsonld)
    assert file_set.name == "foo"
    assert file_set.id == "foo_id"
    assert file_set.description == "bar"
    assert file_set.contained_in == ["some.zip"]
    assert file_set.encoding_format == "application/json"
    assert file_set.excludes == ["*.csv"]
    assert file_set.includes == ["*.json"]
    assert not ctx.issues.errors



"""Script to validate a Croissant JSON-LD file and output errors & warnings if any.

Usage:

```
mlcroissant validate --jsonld /path/to/file.json
```
"""

import sys

from absl import app
from absl import flags
from absl import logging

import mlcroissant as mlc

flags.DEFINE_string(
    "jsonld",
    None,
    "JSON-LD to validate (path to the file or URL).",
    required=True,
)

flags.DEFINE_string(
    "file",
    "",
    "[DEPRECATED] Path to the file to validate.",
    required=False,
)

flags.DEFINE_bool(
    "debug",
    False,
    "Whether to print debug hints.",
)

flags.mark_flag_as_required("jsonld")


FLAGS = flags.FLAGS


def main(argv):
    """Main function launched by the script."""
    del argv
    if FLAGS.file:
        logging.warning("--file is deprecated. Please, use --jsonld with a path or URL")
    jsonld = FLAGS.jsonld or FLAGS.file
    debug = FLAGS.debug
    try:
        mlc.Dataset(jsonld, debug=debug)
        logging.info("Done.")
    except mlc.ValidationError as exception:
        logging.error(exception)
        sys.exit(1)


if __name__ == "__main__":
    app.run(main)



"""Tests for RDF."""

from unittest import mock

from absl import logging
import pytest

from mlcroissant._src.core.context import Context
from mlcroissant._src.core.rdf import make_context
from mlcroissant._src.core.rdf import Rdf
from mlcroissant._src.tests.versions import CONFORMS_TO


@pytest.mark.parametrize(
    "ctx",
    [None] + [Context(conforms_to=conforms_to) for conforms_to in CONFORMS_TO],
)
def test_shorten_value(ctx):
    rdf = Rdf.from_json(
        ctx=ctx,
        json={
            "@context": {
                "bio": "https://bioschemas.org/",
                "replace": "ml:replace",
            }
        },
    )
    assert rdf.shorten_value("https://bioschemas.org/name") == "bio:name"
    # Do not replace the value if it's not a URL...
    assert rdf.shorten_value("ml:replace") == "ml:replace"
    # ...even though the value appears in the context:
    assert "ml:replace" in rdf.reverse_context()


@pytest.mark.parametrize(
    "ctx",
    [None] + [Context(conforms_to=conforms_to) for conforms_to in CONFORMS_TO],
)
def test_shorten_key(ctx):
    rdf = Rdf.from_json(
        ctx=ctx,
        json={
            "@context": {
                "@vocab": "https://base.org/",
                "bio": "https://bioschemas.org/",
                "bar": "bio:bar",
                "baz": {"@id": "bio:baz"},
                "replace": "ml:replace",
            }
        },
    )
    assert rdf.shorten_key("https://base.org/foo") == "foo"
    assert rdf.shorten_key("https://bioschemas.org/name") == "bio:name"
    assert rdf.shorten_key("bio:name") == "bio:name"
    assert rdf.shorten_key("https://bioschemas.org/bar") == "bar"
    assert rdf.shorten_key("bio:bar") == "bar"
    assert rdf.shorten_key("https://bioschemas.org/baz") == "baz"
    assert rdf.shorten_key("bio:baz") == "baz"


@mock.patch.object(logging, "warning")
def test_from_json_without_warning(mock_warning):
    jsonld = {"@context": make_context()}
    rdf = Rdf.from_json(Context(), jsonld)
    assert rdf.context == make_context()
    mock_warning.assert_not_called()


@mock.patch.object(logging, "warning")
def test_from_json_with_warning(mock_warning):
    jsonld = {"@context": {"foo": "bar"}}
    rdf = Rdf.from_json(Context(), jsonld)
    assert rdf.context == {"foo": "bar"}
    mock_warning.assert_called_once()



"""Migration: citation should be citeAs."""


def up(json_ld):
    """Up migration to set conformsTo to croissant specs 1.0."""
    if "citation" in json_ld:
        json_ld["citeAs"] = json_ld["citation"]
        del json_ld["citation"]
    return json_ld



"""Defines the public interface to the `mlcroissant` package."""

from mlcroissant._src import torch
from mlcroissant._src.beam import ReadFromCroissant
from mlcroissant._src.core import constants
from mlcroissant._src.core.constants import DataType
from mlcroissant._src.core.constants import EncodingFormat
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.issues import GenerationError
from mlcroissant._src.core.issues import Issues
from mlcroissant._src.core.issues import ValidationError
from mlcroissant._src.core.rdf import Rdf
from mlcroissant._src.datasets import Dataset
from mlcroissant._src.datasets import Records
from mlcroissant._src.structure_graph.base_node import Node
from mlcroissant._src.structure_graph.nodes.creative_work import CreativeWork
from mlcroissant._src.structure_graph.nodes.field import Field
from mlcroissant._src.structure_graph.nodes.file_object import FileObject
from mlcroissant._src.structure_graph.nodes.file_set import FileSet
from mlcroissant._src.structure_graph.nodes.metadata import Metadata
from mlcroissant._src.structure_graph.nodes.organization import Organization
from mlcroissant._src.structure_graph.nodes.person import Person
from mlcroissant._src.structure_graph.nodes.record_set import RecordSet
from mlcroissant._src.structure_graph.nodes.source import Extract
from mlcroissant._src.structure_graph.nodes.source import FileProperty
from mlcroissant._src.structure_graph.nodes.source import Source
from mlcroissant._src.structure_graph.nodes.source import Transform

__all__ = [
    "constants",
    "Context",
    "CreativeWork",
    "Dataset",
    "DataType",
    "EncodingFormat",
    "Extract",
    "Field",
    "FileObject",
    "FileProperty",
    "FileSet",
    "GenerationError",
    "Issues",
    "Metadata",
    "Node",
    "Organization",
    "Person",
    "Rdf",
    "ReadFromCroissant",
    "Records",
    "RecordSet",
    "Source",
    "torch",
    "Transform",
    "ValidationError",
]
